(*  Title:       Consistency.thy
    Author:      Stephan van Staden, ETH Zurich, Switzerland
    Description: Consistency between the deductive and operational calculi (partial-correctness guarantees).
    Copyright    2012 ETH Zurich
*)

theory Consistency
  imports DeductiveCalculi OperationalCalculi
begin

declare [[ smt_solver = remote_z3]]
declare [[ smt_timeout = 60 ]]
declare [[ z3_options = "-memory:500" ]]

locale consistency_deductive_operational_calculi = full_views_calculus + relationships_between_operational_calculi_for_commands +
  constrains AtomicOperations :: "'b atom \<Rightarrow> bool"
begin
lemma b_triple_kahn_consistent: "v \<lless>P\<ggreater> v' \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<longrightarrow> \<sigma>' \<in> \<lfloor>v'\<rfloor>)"
proof -
  have helper: "\<And>tv v \<sigma>. tv \<in> \<Union>{P. \<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. P = ic_traces_ending_in_state(\<sigma>)} \<longleftrightarrow> (\<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. tv \<in> ic_traces_ending_in_state(\<sigma>))" by auto
  assume "v \<lless>P\<ggreater> v'"
  hence "\<forall>tv \<in> traces_of_view(v). \<forall>tp \<in> P. ic(tv@tp) \<longrightarrow> tv@tp \<in> traces_of_view(v')" by (simp add: b_triple_alt)
  with traces_of_view_def have "\<forall>tv tp. tv \<in> \<Union>{P. \<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. P = ic_traces_ending_in_state(\<sigma>)} \<and> tp \<in> P \<and> ic(tv@tp) \<longrightarrow> tv@tp \<in> \<Union>{P. \<exists>\<sigma>'' \<in> \<lfloor>v'\<rfloor>. P = ic_traces_ending_in_state(\<sigma>'')}" by auto
  with helper have a: "\<forall>tv tp. (\<exists>\<sigma> \<in> \<lfloor>v\<rfloor>. tv \<in> ic_traces_ending_in_state(\<sigma>)) \<and> tp \<in> P \<and> ic(tv@tp) \<longrightarrow> (\<exists>\<sigma>'' \<in> \<lfloor>v'\<rfloor>. tv@tp \<in> ic_traces_ending_in_state(\<sigma>''))" by auto
  show "\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<longrightarrow> \<sigma>' \<in> \<lfloor>v'\<rfloor>"
  proof (rule ballI)
    fix \<sigma>
    assume b: "\<sigma> \<in> \<lfloor>v\<rfloor>"
    show "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<longrightarrow> \<sigma>' \<in> \<lfloor>v'\<rfloor>"
    proof (rule impI)
      assume "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
      hence "\<exists>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'}" by (simp add: d_kahn_def)
      hence "\<exists>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). t' \<in> {t};P" by (metis a_kahn_def bottom_def insertI1 subsetD)
      then obtain t t' where c: "t \<in> ic_traces_ending_in_state(\<sigma>)" and d: "t' \<in> ic_traces_ending_in_state(\<sigma>')" and e: "t' \<in> {t};P" by blast
      from d have f: "ic(t')" by (auto simp add: ic_traces_ending_in_state_def)
      from e have "\<exists>tp \<in> P. t' = t@tp" by (metis elem_of_lang_concat singleton_iff)
      then obtain tp where g: "tp \<in> P" and h: "t' = t@tp" by blast
      from f h have i: "ic(t@tp)" by simp
      from a b c d g h i have "\<exists>\<sigma>'' \<in> \<lfloor>v'\<rfloor>. t@tp \<in> ic_traces_ending_in_state(\<sigma>'')" by metis
      with d h show "\<sigma>' \<in> \<lfloor>v'\<rfloor>" by (metis trace_can_end_only_in_one_state)
    qed
  qed
qed
lemma b_triple_state_trans_consistent: "v \<lless>P\<ggreater> v' \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. \<lbrakk>P\<rbrakk>\<^isub>d(\<sigma>) \<subseteq> \<lfloor>v'\<rfloor>)"
  by (metis b_triple_kahn_consistent d_state_transformation_function_def mem_def Collect_def subsetI)
lemma b_triple_plotkin_consistent: "v \<lless>P\<ggreater> v' \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. <P,\<sigma>>\<rightarrow>\<^isub>d\<^sup>\<star><skip,\<sigma>'> \<longrightarrow> \<sigma>' \<in> \<lfloor>v'\<rfloor>)"
  by (metis b_triple_kahn_consistent plotkin_kahn_descr_rel)
lemma f_triple_kahn_consistent: "v \<langle>P\<rangle> v' \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<longrightarrow> \<sigma>' \<in> \<lfloor>v'\<rfloor>)"
  by (metis b_triple_kahn_consistent f_triple_b_triple_relationship)
lemma f_triple_state_trans_consistent: "v \<langle>P\<rangle> v' \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. \<lbrakk>P\<rbrakk>\<^isub>d(\<sigma>) \<subseteq> \<lfloor>v'\<rfloor>)"
  by (metis f_triple_kahn_consistent d_state_transformation_function_def mem_def Collect_def subsetI)
lemma f_triple_plotkin_consistent: "v \<langle>P\<rangle> v' \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. <P,\<sigma>>\<rightarrow>\<^isub>d\<^sup>\<star><skip,\<sigma>'> \<longrightarrow> \<sigma>' \<in> \<lfloor>v'\<rfloor>)"
  by (metis f_triple_kahn_consistent plotkin_kahn_descr_rel)
lemma views_triple_kahn_consistent: "\<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. <C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>' \<longrightarrow> \<sigma>' \<in> \<lfloor>v'\<rfloor>)"
  by (metis f_triple_kahn_consistent c_kahn_def views_triple_f_triple_relationship)
lemma views_triple_state_trans_consistent: "\<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. \<lbrakk>C\<rbrakk>\<^isub>c(\<sigma>) \<subseteq> \<lfloor>v'\<rfloor>)"
  by (metis views_triple_kahn_consistent c_state_transformation_function_def mem_def Collect_def subsetI)
lemma views_triple_plotkin_consistent: "\<lbrace>v\<rbrace> C \<lbrace>v'\<rbrace> \<Longrightarrow> (\<forall>\<sigma> \<in> \<lfloor>v\<rfloor>. <C,\<sigma>>\<rightarrow>\<^isub>c\<^sup>\<star><skip,\<sigma>'> \<longrightarrow> \<sigma>' \<in> \<lfloor>v'\<rfloor>)"
  by (metis views_triple_kahn_consistent plotkin_kahn_comm_rel)
end

end
