(*  Title:       Interleaving.thy
    Author:      Stephan van Staden, ETH Zurich, Switzerland
    Description: List interleaving.
    Copyright    2012 ETH Zurich
*)

theory Interleaving
  imports Main
begin

declare [[ smt_solver = remote_z3]]
declare [[ smt_timeout = 60 ]]
declare [[ z3_options = "-memory:500" ]]

fun interleave :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list set" (infix "\<otimes>" 65) where
   "[] \<otimes> q = {q}"
 | "p \<otimes> [] = {p}"
 | "(e#p) \<otimes> (e'#q) = {r'. \<exists>r \<in> p \<otimes> (e'#q). r' = e#r} \<union> {r'. \<exists>r \<in> (e#p) \<otimes> q. r' = e'#r}"

lemma elem_interleave_with_empty1: "p \<in> []\<otimes>q \<longleftrightarrow> p=q"
  by (metis interleave.simps(1) singleton_iff)

lemma elem_interleave_with_empty2: "p \<in> q\<otimes>[] \<longleftrightarrow> p=q"
  by (metis elem_interleave_with_empty1 interleave.simps(1) interleave.simps(2) neq_Nil_conv)

lemma interleave_contains_empty: "[] \<in> p \<otimes> q \<longleftrightarrow> p = [] \<and> q = []"
proof -
have interleave_contains_empty_helper: "[] \<in> p \<otimes> q \<Longrightarrow> p = []"
proof -
  assume a: "[] \<in> p \<otimes> q"
  show "p = []"
  proof (cases p)
    case Nil then show ?thesis by (metis Nil)
  next
    case (Cons e p') then
    have False
    proof (cases q)
      case Nil then show ?thesis by (metis Collect_def Cons a empty_def insert_code interleave.simps(2) list.simps(3) mem_def)
    next
      case (Cons e' q') then
      have b: "p \<otimes> q = {r'. \<exists>r \<in> p' \<otimes> (e'#q'). r' = e#r} \<union> {r'. \<exists>r \<in> (e#p') \<otimes> q'. r' = e'#r}" by (metis `p = e # p'` interleave.simps(3))
      have "\<And>r'. r' \<in> p \<otimes> q \<Longrightarrow> r' \<noteq> []"
      proof -
        fix r' assume c: "r' \<in> p \<otimes> q"
        from b c have "(\<exists>r \<in> p' \<otimes> (e'#q'). r' = e#r) \<or> (\<exists>r \<in> (e#p') \<otimes> q'. r' = e'#r)" by simp
        then show "r' \<noteq> []" by (metis list.simps(2))
      qed
      then show ?thesis by (metis a)
    qed
    then show ?thesis ..
  qed
qed
show ?thesis by (metis interleave.simps(1) interleave_contains_empty_helper singleton_iff)
qed

lemma interleave_contains_nonempty: "e#r \<in> p \<otimes> q \<longleftrightarrow> (\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> q) \<or> (\<exists>q'. q = e#q' \<and> r \<in> p \<otimes> q')"
proof -
have interleave_contains_nonempty_helper1: "e#r \<in> p \<otimes> q \<Longrightarrow> (\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> q) \<or> (\<exists>q'. q = e#q' \<and> r \<in> p \<otimes> q')"
proof -
  assume a: "e#r \<in> p \<otimes> q"
  show "(\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> q) \<or> (\<exists>q'. q = e#q' \<and> r \<in> p \<otimes> q')"
  proof (cases p)
    case Nil then
      have "p \<otimes> q = {q}" by simp
      from a have "\<exists>q'. q = e#q' \<and> r \<in> p \<otimes> q'" by (metis Nil emptyE insertE insertI1 interleave.simps(1))
      then show ?thesis by metis
  next
    case (Cons e' p')
      have ?thesis
      proof (cases q)
        case Nil then
          have "p \<otimes> q = {p}" by (metis Cons interleave.simps(2))
          from a have "\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> q" by (metis Nil `p \<otimes> q = {p}` interleave.simps(1) interleave.simps(2) neq_Nil_conv singleton_iff)
          then show ?thesis by metis
      next
        case (Cons eq q')
          have b: "p \<otimes> q = {r'. \<exists>r \<in> p' \<otimes> (eq#q'). r' = e'#r} \<union> {r'. \<exists>r \<in> (e'#p') \<otimes> q'. r' = eq#r}" by (metis Cons `p = e' # p'` interleave.simps(3))
          from a b have "(\<exists>r' \<in> p' \<otimes> (eq#q'). e#r = e'#r') \<or> (\<exists>r' \<in> (e'#p') \<otimes> q'. e#r = eq#r')" by simp
          then show ?thesis by (metis Cons `p = e' # p'` list.inject)
      qed
      then show ?thesis by simp
  qed
qed
have interleave_contains_nonempty_helper2: "(\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> q) \<or> (\<exists>q'. q = e#q' \<and> r \<in> p \<otimes> q') \<Longrightarrow> e#r \<in> p \<otimes> q"
proof -
  assume a: "(\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> q) \<or> (\<exists>q'. q = e#q' \<and> r \<in> p \<otimes> q')"
  show "e#r \<in> p \<otimes> q"
  proof (cases q)
    case Nil then
    have "(\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> []) \<or> (\<exists>q'. [] = e#q' \<and> r \<in> p \<otimes> q')" by (metis a)
    then have "\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> []" by (metis list.simps(2))
    then have "\<exists>p'. p = e#p' \<and> r \<in> {p'}" by (metis interleave.simps(1) interleave.simps(2) neq_Nil_conv)
    then have "p = e#r" by (metis singleton_iff)
    then show ?thesis by (metis Nil insert_code interleave.simps(2) mem_def)
  next
    case (Cons eq q')
    from a have b: "(\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> (eq#q')) \<or> (\<exists>q''. eq#q' = e#q'' \<and> r \<in> p \<otimes> q'')" by (metis Cons list.inject)
    have c: "\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> (eq#q') \<Longrightarrow> e#r \<in> p \<otimes> q"
    proof -
      assume d: "\<exists>p'. p = e#p' \<and> r \<in> p' \<otimes> (eq#q')"
      from d obtain p' where f: "p = e#p'" and g: "r \<in> p' \<otimes> (eq#q')" by blast
      have h: "(e#p') \<otimes> (eq#q') = {r'. \<exists>r \<in> p' \<otimes> (eq#q'). r' = e#r} \<union> {r'. \<exists>r \<in> (e#p') \<otimes> q'. r' = eq#r}" by (metis interleave.simps(3))
      from g have i: "e#r \<in> {r'. \<exists>r \<in> p' \<otimes> (eq#q'). r' = e#r}" by (simp add: Collect_def mem_def)
      from h i have "e#r \<in> (e#p') \<otimes> (eq#q')" by (simp add: UnCI)
      show ?thesis by (metis Cons `e # r \<in> (e # p') \<otimes> (eq # q')` f)
    qed
    have j: "\<exists>q''. eq#q' = e#q'' \<and> r \<in> p \<otimes> q'' \<Longrightarrow> e#r \<in> p \<otimes> q"
    proof -
      assume k: "\<exists>q''. eq#q' = e#q'' \<and> r \<in> p \<otimes> q''"
      then have l: "eq = e \<and> r \<in> p \<otimes> q'" by (metis hd.simps tl.simps(2))
      show "e#r \<in> p \<otimes> q"
      proof (cases p)
        assume p_nil: "p = []"
        from p_nil l have r_is_q': "eq = e \<and> r = q'" by simp
        then have "p \<otimes> q = {e#r}" by (metis Cons interleave.simps(1) p_nil)
        then show ?thesis by (metis singleton_iff)
      next
        fix ep p' assume p_cons: "p = ep#p'"
        have m: "p \<otimes> q = {r'. \<exists>r \<in> p' \<otimes> (eq#q'). r' = ep#r} \<union> {r'. \<exists>r \<in> (ep#p') \<otimes> q'. r' = eq#r}" by (metis Cons interleave.simps(3) p_cons)
        from l have n: "e#r \<in> {r'. \<exists>r \<in> (ep#p') \<otimes> q'. r' = eq#r}" by (simp add: Collect_def mem_def p_cons)
        from m n show ?thesis by (simp add: UnCI)
      qed
    qed
    from c j show ?thesis by (metis b)
  qed
qed
show ?thesis by (metis interleave_contains_nonempty_helper1 interleave_contains_nonempty_helper2)
qed

lemma interleave_commutative: "p \<otimes> q = q \<otimes> p"
proof -
have interleave_commutative_helper: "\<And>q. p \<otimes> q = q \<otimes> p"
proof (induct p)
  show "\<And>q. [] \<otimes> q = q \<otimes> []" by (metis interleave.simps(1) interleave.simps(2) neq_Nil_conv)
next
  fix a p q assume IH1: "\<And>q. p \<otimes> q = q \<otimes> p"
  show "(a # p) \<otimes> q = q \<otimes> (a # p)"
  proof (induct q)
    show "(a # p) \<otimes> [] = [] \<otimes> (a # p)" by (metis interleave.simps(1) interleave.simps(2))
  next
    fix aa q' assume IH2: "(a # p) \<otimes> q' = q' \<otimes> (a # p)"
    show "(a # p) \<otimes> (aa # q') = (aa # q') \<otimes> (a # p)"
      proof -
        have "(a # p) \<otimes> (aa # q') = {r'. \<exists>r \<in> p \<otimes> (aa#q'). r' = a#r} \<union> {r'. \<exists>r \<in> (a#p) \<otimes> q'. r' = aa#r}" by (metis interleave.simps(3))
        also have "\<dots> = {r'. \<exists>r \<in> (a#p) \<otimes> q'. r' = aa#r} \<union> {r'. \<exists>r \<in> p \<otimes> (aa#q'). r' = a#r}" by (metis Un_commute)
        also have "\<dots> = {r'. \<exists>r \<in> q' \<otimes> (a#p). r' = aa#r} \<union> {r'. \<exists>r \<in> p \<otimes> (aa#q'). r' = a#r}" by (simp add: IH2)
        also have "\<dots> = {r'. \<exists>r \<in> q' \<otimes> (a#p). r' = aa#r} \<union> {r'. \<exists>r \<in> (aa#q') \<otimes> p. r' = a#r}" by (simp add: IH1)
        also have "\<dots> = (aa # q') \<otimes> (a # p)" by (metis interleave.simps(3))
        finally show ?thesis .
      qed
  qed
qed
show ?thesis by (simp add: interleave_commutative_helper)
qed

lemma interleave_assoc: "(\<exists>s \<in> p \<otimes> q. t \<in> s \<otimes> r) \<longleftrightarrow> (\<exists>u \<in> q \<otimes> r. t \<in> p \<otimes> u)"
proof -
have interleave_assoc_helper: "\<And>p q r s. s \<in> p \<otimes> q \<Longrightarrow> t \<in> s \<otimes> r \<Longrightarrow> (\<exists>u \<in> q \<otimes> r. t \<in> p \<otimes> u)"
proof (induct t)
  fix p q r s
  assume a: "s \<in> p \<otimes> q"
  assume b: "[] \<in> s \<otimes> r"
  from b have c: "s = [] \<and> r = []" by (metis interleave_contains_empty)
  from a c have d: "p = [] \<and> q = []" by (metis interleave_contains_empty)
  from a b c d show "\<exists>u \<in> q \<otimes> r. [] \<in> p \<otimes> u" by simp
next
  fix a::'a and t::"'a list" and p::"'a list" and q r s
  assume IH: "(\<And>p q r s. s \<in> p \<otimes> q \<Longrightarrow> t \<in> s \<otimes> r \<Longrightarrow> \<exists>u\<in>q \<otimes> r. t \<in> p \<otimes> u)"
  assume e: "s \<in> p \<otimes> q"
  assume f: "a#t \<in> s\<otimes>r"
  show "\<exists>u\<in>q \<otimes> r. a # t \<in> p \<otimes> u"
  proof -
    from f have g: "(\<exists>p'. s = a#p' \<and> t \<in> p' \<otimes> r) \<or> (\<exists>q'. r = a#q' \<and> t \<in> s \<otimes> q')" by (simp add: interleave_contains_nonempty)
    have h: "(\<exists>p'. s = a#p' \<and> t \<in> p' \<otimes> r) \<Longrightarrow> ?thesis"
    proof -
      assume "\<exists>p'. s = a#p' \<and> t \<in> p' \<otimes> r"
      then obtain p' where i: "s = a#p'" and j: "t \<in> p' \<otimes> r" by blast
      from i e have k: "(\<exists>p''. p = a#p'' \<and> p' \<in> p'' \<otimes> q) \<or> (\<exists>q'. q = a#q' \<and> p' \<in> p \<otimes> q')" by (simp add: interleave_contains_nonempty)
      have l: "(\<exists>p''. p = a#p'' \<and> p' \<in> p'' \<otimes> q) \<Longrightarrow> ?thesis"
      proof -
        assume "\<exists>p''. p = a#p'' \<and> p' \<in> p'' \<otimes> q"
        then obtain p'' where a1: "p = a#p''" and a2: "p' \<in> p'' \<otimes> q" by blast
        from a2 j have "\<exists>u\<in>q \<otimes> r. t \<in> p'' \<otimes> u" by (rule IH)
        then show ?thesis by (metis a1 interleave_contains_nonempty)
      qed
      have m: "(\<exists>q'. q = a#q' \<and> p' \<in> p \<otimes> q') \<Longrightarrow> ?thesis"
      proof -
        assume "\<exists>q'. q = a#q' \<and> p' \<in> p \<otimes> q'"
        then obtain q' where n: "q = a#q'" and o: "p' \<in> p \<otimes> q'" by blast
        from o j have f4: "\<exists>u\<in>q' \<otimes> r. t \<in> p \<otimes> u" by (rule IH)
        then show ?thesis by (metis n interleave_contains_nonempty)
      qed
      from l m show ?thesis by (metis k)
    qed
    have q: "(\<exists>q'. r = a#q' \<and> t \<in> s \<otimes> q') \<Longrightarrow> ?thesis"
    proof -
      assume "\<exists>q'. r = a#q' \<and> t \<in> s \<otimes> q'"
      then obtain q' where r: "r = a#q'" and s: "t \<in> s \<otimes> q'" by blast
      from e s have "\<exists>u\<in>q \<otimes> q'. t \<in> p \<otimes> u" by (rule IH)
      then show "\<exists>v\<in>q \<otimes> r. a # t \<in> p \<otimes> v" by (metis r interleave_contains_nonempty)
    qed
    from h q show ?thesis by (metis g)
  qed
qed
show ?thesis by (metis interleave_assoc_helper interleave_commutative)
qed

lemma set_interleave: "\<Union>{S. \<exists>l'' \<in> l\<otimes>l'. S = set l''} = set l \<union> set l'"
proof -
  have helper: "\<And>l' e. (\<exists>l'' \<in> l\<otimes>l'. e \<in> set l'') \<longleftrightarrow> e \<in> set l \<or> e \<in> set l'"
  proof (induct l)
    fix l' e
    show "(\<exists>l''\<in>[] \<otimes> l'. e \<in> set l'') \<longleftrightarrow> (e \<in> set [] \<or> e \<in> set l')" by auto
  next
    fix a l l' e
    assume IH1: "\<And>l' e. (\<exists>l''\<in>l \<otimes> l'. e \<in> set l'') \<longleftrightarrow> (e \<in> set l \<or> e \<in> set l')"
    show "(\<exists>l''\<in>(a # l) \<otimes> l'. e \<in> set l'') \<longleftrightarrow> (e \<in> set (a # l) \<or> e \<in> set l')"
    proof (induct l')
      show "(\<exists>l''\<in>(a # l) \<otimes> []. e \<in> set l'') \<longleftrightarrow> (e \<in> set (a # l) \<or> e \<in> set [])" by auto
    next
      fix aa l'
      assume IH2: "(\<exists>l''\<in>(a # l) \<otimes> l'. e \<in> set l'') \<longleftrightarrow> (e \<in> set (a # l) \<or> e \<in> set l')"
      show "(\<exists>l''\<in>(a # l) \<otimes> (aa # l'). e \<in> set l'') \<longleftrightarrow> (e \<in> set (a # l) \<or> e \<in> set (aa # l'))"
      proof -
        have "(\<exists>l''\<in>(a # l) \<otimes> (aa # l'). e \<in> set l'') \<longleftrightarrow> (\<exists>l''. (\<exists>r \<in> l \<otimes> (aa#l'). l'' = a#r \<and> e \<in> set l'') \<or> (\<exists>r \<in> (a#l) \<otimes> l'. l'' = aa#r \<and> e \<in> set l''))" by auto
        also have "\<dots> \<longleftrightarrow> (\<exists>r \<in> l \<otimes> (aa#l'). e = a \<or> e \<in> set r) \<or> (\<exists>r \<in> (a#l) \<otimes> l'. e = aa \<or> e \<in> set r)" by auto
        also have "\<dots> \<longleftrightarrow> (e = a \<or> (\<exists>r \<in> l \<otimes> (aa#l'). e \<in> set r)) \<or> (e = aa \<or> (\<exists>r \<in> (a#l) \<otimes> l'. e \<in> set r))" by (metis IH1 mem_def member_rec(1) member_set)
        also have "\<dots> \<longleftrightarrow> e = a \<or> e \<in> set l \<or> e \<in> set (aa#l') \<or> e = aa \<or> (\<exists>r \<in> (a#l) \<otimes> l'. e \<in> set r)" by (metis (full_types) IH1)
        also have "\<dots> \<longleftrightarrow> e = a \<or> e \<in> set l \<or> e \<in> set (aa#l') \<or> e = aa \<or> e \<in> set (a#l) \<or> e \<in> set l'" by (metis (full_types) IH2)
        finally show "(\<exists>l''\<in>(a # l) \<otimes> (aa # l'). e \<in> set l'') \<longleftrightarrow> (e \<in> set (a # l) \<or> e \<in> set (aa # l'))" by auto
      qed
   qed
  qed
  have "\<forall>e. e \<in> \<Union>{S. \<exists>l'' \<in> l\<otimes>l'. S = set l''} \<longleftrightarrow> e \<in> set l \<or> e \<in> set l'"
  proof (rule allI)
    fix e
    have "e \<in> \<Union>{S. \<exists>l'' \<in> l\<otimes>l'. S = set l''} \<longleftrightarrow> (\<exists>l'' \<in> l\<otimes>l'. e \<in> set l'')" by auto
    also have "\<dots> \<longleftrightarrow> e \<in> set l \<or> e \<in> set l'" by (rule helper)
    finally show "e \<in> \<Union>{S. \<exists>l'' \<in> l\<otimes>l'. S = set l''} \<longleftrightarrow> e \<in> set l \<or> e \<in> set l'" by simp
  qed
  then show ?thesis by auto
qed

end
