(*  Title:       OperationalCalculi.thy
    Author:      Stephan van Staden, ETH Zurich, Switzerland
    Description: The various operational calculi for descriptions and commands.
    Copyright    2012 ETH Zurich
*)

theory OperationalCalculi
  imports AbstractCalculi DescriptionsAndCommands
begin

declare [[ smt_solver = remote_z3]]
declare [[ smt_timeout = 60 ]]
declare [[ z3_options = "-memory:500" ]]

locale atomic_operations = 
  fixes AtomicOperations :: "'a atom set"

locale small_step_calculi_for_descriptions = atomic_operations
begin
definition Actions :: "'a description set" where
  "Actions \<equiv> {skip} \<union> Rep_atom ` AtomicOperations"
end

sublocale small_step_calculi_for_descriptions < a!:abstract_small_step_calculi Actions
apply (rule abstract_small_step_calculi.intro)
by (metis Actions_def Un_commute Un_insert_right insert_subset order_eq_refl)

context small_step_calculi_for_descriptions
begin
abbreviation (input) a_plotkin :: "'a description \<Rightarrow> 'a description \<Rightarrow> 'a description \<Rightarrow> 'a description \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>a<_,_>" [54,54,54,54] 53) where
  "<P,s>\<rightarrow>\<^isub>a<P',s'> \<equiv> a.a_plotkin P s P' s'"
definition d_plotkin :: "'a description \<Rightarrow> 'a \<Rightarrow> 'a description \<Rightarrow> 'a \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>d<_,_>" [54,54,54,54] 53) where
  "<P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'> \<equiv> (\<exists>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a<P',{t'}>)"
lemma d_plotkin_alt_def: "<P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'> \<longleftrightarrow> (\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a<P',{t'}>)"
proof -
  have part1: "(\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a<P',{t'}>) \<longrightarrow> <P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'>"
    by (metis d_plotkin_def ex_in_conv ic_traces_ending_in_state_not_empty)
  have part2: "<P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'> \<longrightarrow> (\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a<P',{t'}>)"
  proof (rule impI)
    assume a: "<P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'>"
    then have "\<exists>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). \<exists>Q. Q \<in> Actions \<and> P \<supseteq> Q;P' \<and> {t};Q \<supseteq> {t'}" by (simp add: d_plotkin_def a.a_plotkin_def)
    then obtain t t' Q where b: "t \<in> ic_traces_ending_in_state(\<sigma>)" and c: "t' \<in> ic_traces_ending_in_state(\<sigma>')" and d: "Q \<in> Actions \<and> P \<supseteq> Q;P'" and "{t'} \<subseteq> {t};Q" by blast
    then have "\<exists>tq \<in> Q. t' = t@tq" by (metis elem_of_lang_concat insert_subset singletonE)
    then obtain tq where e: "tq \<in> Q" and f: "t' = t@tq" by blast
    from f c have g: "t@tq \<in> ic_traces_ending_in_state(\<sigma>')" by simp
    have h: "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a<P',{t'}>"
    proof (rule ballI)
      fix t1
      assume i: "t1 \<in> ic_traces_ending_in_state(\<sigma>)"
      from b i g have j: "t1@tq \<in> ic_traces_ending_in_state(\<sigma>')" by (rule history_independence)
      have k: "<P,{t1}>\<rightarrow>\<^isub>a<P',{t1@tq}>" by (unfold a.a_plotkin_def, auto simp add: elem_of_lang_concat singletonE, metis d e)
      from j k show "\<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t1}>\<rightarrow>\<^isub>a<P',{t'}>" by auto
    qed
    from a h show "(\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a<P',{t'}>)" by (simp add: d_plotkin_def)
  qed
  from part1 part2 show ?thesis by metis
qed
lemma PDatom: "a \<in> AtomicOperations \<Longrightarrow> \<sigma>' \<in> a of_state \<sigma> \<Longrightarrow> <Rep_atom a,\<sigma>>\<rightarrow>\<^isub>d<skip,\<sigma>'>"
proof -
  assume "a \<in> AtomicOperations"
  then have a: "Rep_atom a \<in> Actions" by (metis Actions_def image_eqI insertCI insert_is_Un)
  assume "\<sigma>' \<in> a of_state \<sigma>"
  hence "[(\<sigma>,\<sigma>')] \<in> Rep_atom(a)" by (metis (no_types) Collect_def atom_applied_to_state_def mem_def) 
  hence "{[(\<sigma>,\<sigma>),(\<sigma>,\<sigma>')]} \<subseteq> {[(\<sigma>,\<sigma>)]};Rep_atom(a)" by (metis (no_types) empty_subsetI append_Cons elem_of_lang_concat insertI1 singletonE insert_subset lang_concat_unit1 skip_def)
  with a have "<Rep_atom a,{[(\<sigma>,\<sigma>)]}>\<rightarrow>\<^isub>a<skip,{[(\<sigma>,\<sigma>),(\<sigma>,\<sigma>')]}>" by (rule a.Paction)
  thus "<Rep_atom a,\<sigma>>\<rightarrow>\<^isub>d<skip,\<sigma>'>" by (metis d_plotkin_def trivial_ic_trace_ending_in_state trivial_ic_trace_ending_in_state_length2)
qed
lemma PDseq1: "<skip;P,\<sigma>>\<rightarrow>\<^isub>d<P,\<sigma>>"
  by (metis a.Pseq1 all_not_in_conv d_plotkin_def ic_traces_ending_in_state_not_empty)
lemma PDseq2: "<P,\<sigma>>\<rightarrow>\<^isub>d<R,\<sigma>'> \<Longrightarrow> <P;P',\<sigma>>\<rightarrow>\<^isub>d<R;P',\<sigma>'>"
  by (metis (no_types) a.Pseq2 d_plotkin_def)
lemma PDchoice: "P \<in> X \<Longrightarrow> <\<Union>X,\<sigma>>\<rightarrow>\<^isub>d<P,\<sigma>>"
  by (metis a.Pchoice d_plotkin_def all_not_in_conv ic_traces_ending_in_state_not_empty)
lemma PDfuturechoice: "<P;(P'\<union>P''),\<sigma>>\<rightarrow>\<^isub>d<P;P',\<sigma>>"
proof -
  have Pfuturechoice: "\<And>s P P' P''. <P;(P'\<union>P''),s>\<rightarrow>\<^isub>a<P;P',s>"
    by (unfold a.a_plotkin_def, metis Un_upper1 a.skip_in_Actions lang_concat_distr_union2 lang_concat_unit1 lang_concat_unit2 order_eq_refl)
  show "<P;(P'\<union>P''),\<sigma>>\<rightarrow>\<^isub>d<P;P',\<sigma>>" by (metis Pfuturechoice all_not_in_conv d_plotkin_def ic_traces_ending_in_state_not_empty)
qed
lemma PDiter1: "<P\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>d<skip,\<sigma>>"
  by (metis a.Piter1 all_not_in_conv d_plotkin_def ic_traces_ending_in_state_not_empty)
lemma PDiter2: "<P\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>d<P;P\<^sup>\<star>,\<sigma>>"
  by (metis a.Piter2 all_not_in_conv d_plotkin_def ic_traces_ending_in_state_not_empty)
lemma PDconc1: "<skip\<parallel>P,\<sigma>>\<rightarrow>\<^isub>d<P,\<sigma>>" 
  by (metis a.Pconc1 all_not_in_conv d_plotkin_def ic_traces_ending_in_state_not_empty)
lemma PDconc2: "<P\<parallel>skip,\<sigma>>\<rightarrow>\<^isub>d<P,\<sigma>>"
  by (metis PDconc1 lang_interleave_commutative)
lemma PDconc3: "<P,\<sigma>>\<rightarrow>\<^isub>d<R,\<sigma>'> \<Longrightarrow> <P\<parallel>P',\<sigma>>\<rightarrow>\<^isub>d<R\<parallel>P',\<sigma>'>"
  by (metis (no_types) a.Pconc3 d_plotkin_def)
lemma PDconc4: "<P,\<sigma>>\<rightarrow>\<^isub>d<R,\<sigma>'> \<Longrightarrow> <P'\<parallel>P,\<sigma>>\<rightarrow>\<^isub>d<P'\<parallel>R,\<sigma>'>"
  by (metis PDconc3 lang_interleave_commutative)
fun d_plotkin_iter :: "'a description \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a description \<Rightarrow> 'a \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>d[_]<_,_>" [54,54,54,54,54] 53) where
   "<P,\<sigma>>\<rightarrow>\<^isub>d[0]<P',\<sigma>'> = (P = P' \<and> \<sigma> = \<sigma>')"
 | "<P,\<sigma>>\<rightarrow>\<^isub>d[Suc n]<P',\<sigma>'> = (\<exists>P'' \<sigma>''. <P,\<sigma>>\<rightarrow>\<^isub>d<P'',\<sigma>''> \<and> <P'',\<sigma>''>\<rightarrow>\<^isub>d[n]<P',\<sigma>'>)"
definition d_plotkin_refl_trans :: "'a description \<Rightarrow> 'a \<Rightarrow> 'a description \<Rightarrow> 'a \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>d\<^sup>\<star><_,_>" [54,54,54,54] 53) where
   "<P,\<sigma>>\<rightarrow>\<^isub>d\<^sup>\<star><P',\<sigma>'> \<equiv> (\<exists>n. <P,\<sigma>>\<rightarrow>\<^isub>d[n]<P',\<sigma>'>)"

abbreviation (input) a_milner :: "'a description \<Rightarrow> 'a description \<Rightarrow> 'a description \<Rightarrow> bool" ("_ -_\<rightarrow>\<^isub>a _" [54,54,54] 53) where
  "P -Q\<rightarrow>\<^isub>a R \<equiv> a.a_milner P Q R"
end


definition d_kahn :: "'a description \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>d_" [54,54,54] 53) where
  "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<equiv> (\<exists>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'})"
lemma d_kahn_alt_def: "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<longleftrightarrow> (\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'})"
proof -
  have part1: "(\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'}) \<longrightarrow> <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
    by (metis d_kahn_def ex_in_conv ic_traces_ending_in_state_not_empty)
  have part2: "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<longrightarrow> (\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'})"
  proof (rule impI)
    assume a: "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
    then have "\<exists>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). {t'} \<subseteq> {t};P" by (simp add: d_kahn_def a_kahn_def)
    then obtain t t' where b: "t \<in> ic_traces_ending_in_state(\<sigma>)" and c: "t' \<in> ic_traces_ending_in_state(\<sigma>')" and "{t'} \<subseteq> {t};P" by blast
    then have "\<exists>tp \<in> P. t' = t@tp" by (metis elem_of_lang_concat insert_subset singletonE)
    then obtain tp where d: "tp \<in> P" and e: "t' = t@tp" by blast
    from e c have f: "t@tp \<in> ic_traces_ending_in_state(\<sigma>')" by simp
    have g: "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'}"
    proof (rule ballI)
      fix t1
      assume h: "t1 \<in> ic_traces_ending_in_state(\<sigma>)"
      from b h f have i: "t1@tp \<in> ic_traces_ending_in_state(\<sigma>')" by (rule history_independence)
      have j: "<P,{t1}>\<rightarrow>\<^isub>a{t1@tp}" by (unfold a_kahn_def, auto simp add: elem_of_lang_concat singletonE d)
      from i j show "\<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t1}>\<rightarrow>\<^isub>a{t'}" by auto
    qed
    from a g show "(\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'})" by (simp add: d_kahn_def)
  qed
  from part1 part2 show ?thesis by metis
qed
theorem KDatom: "\<sigma>' \<in> a of_state \<sigma> \<Longrightarrow> <Rep_atom a,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
proof -
  assume "\<sigma>' \<in> a of_state \<sigma>"
  hence "[(\<sigma>,\<sigma>')] \<in> Rep_atom(a)" by (metis (no_types) Collect_def atom_applied_to_state_def mem_def) 
  hence "{[(\<sigma>,\<sigma>),(\<sigma>,\<sigma>')]} \<subseteq> {[(\<sigma>,\<sigma>)]};Rep_atom(a)" by (metis (no_types) empty_subsetI append_Cons elem_of_lang_concat insertI1 singletonE insert_subset lang_concat_unit1 skip_def)
  hence "<Rep_atom a,{[(\<sigma>,\<sigma>)]}>\<rightarrow>\<^isub>a{[(\<sigma>,\<sigma>),(\<sigma>,\<sigma>')]}" by (metis a_kahn_def)
  thus "<Rep_atom a,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by (metis d_kahn_def trivial_ic_trace_ending_in_state trivial_ic_trace_ending_in_state_length2)
qed
theorem KDskip: "<skip,\<sigma>>\<rightarrow>\<^isub>d\<sigma>"
  by (metis Kskip d_kahn_alt_def)
theorem KDseq: "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<Longrightarrow> <P',\<sigma>'>\<rightarrow>\<^isub>d\<sigma>'' \<Longrightarrow> <P;P',\<sigma>>\<rightarrow>\<^isub>d\<sigma>''"
proof -
  assume "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
  then have a: "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'}" by (metis d_kahn_alt_def)
  assume "<P',\<sigma>'>\<rightarrow>\<^isub>d\<sigma>''"
  then have b: "\<forall>t' \<in> ic_traces_ending_in_state(\<sigma>'). \<exists>t'' \<in> ic_traces_ending_in_state(\<sigma>''). <P',{t'}>\<rightarrow>\<^isub>a{t''}" by (metis d_kahn_alt_def)
  from a b Kseq have "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t'' \<in> ic_traces_ending_in_state(\<sigma>''). <P;P',{t}>\<rightarrow>\<^isub>a{t''}" by metis
  then show "<P;P',\<sigma>>\<rightarrow>\<^isub>d\<sigma>''" by (metis d_kahn_alt_def)
qed
theorem KDchoice: "P \<in> X \<Longrightarrow> <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<Longrightarrow> <\<Union>X,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
proof -
  assume a: "P \<in> X"
  assume "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
  then have b: "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'}" by (metis d_kahn_alt_def)
  show "<\<Union>X,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by (unfold d_kahn_alt_def, metis Kchoice a b)
qed
theorem KDiter1: "<P\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>d\<sigma>"
  by (metis Kiter1 d_kahn_alt_def)
theorem KDiter2: "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<Longrightarrow> <P\<^sup>\<star>,\<sigma>'>\<rightarrow>\<^isub>d\<sigma>'' \<Longrightarrow> <P\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>d\<sigma>''"
proof -
  assume "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
  then have a: "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'}" by (metis d_kahn_alt_def)
  assume "<P\<^sup>\<star>,\<sigma>'>\<rightarrow>\<^isub>d\<sigma>''"
  then have b: "\<forall>t' \<in> ic_traces_ending_in_state(\<sigma>'). \<exists>t'' \<in> ic_traces_ending_in_state(\<sigma>''). <P\<^sup>\<star>,{t'}>\<rightarrow>\<^isub>a{t''}" by (metis d_kahn_alt_def)
  from a b Kiter2 have "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t'' \<in> ic_traces_ending_in_state(\<sigma>''). <P\<^sup>\<star>,{t}>\<rightarrow>\<^isub>a{t''}" by (metis (no_types))
  then show "<P\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>d\<sigma>''" by (metis d_kahn_alt_def)
qed
theorem KDconc1: "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<Longrightarrow> <P',\<sigma>'>\<rightarrow>\<^isub>d\<sigma>'' \<Longrightarrow> <P\<parallel>P',\<sigma>>\<rightarrow>\<^isub>d\<sigma>''"
proof -
  assume "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
  then have a: "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). <P,{t}>\<rightarrow>\<^isub>a{t'}" by (metis d_kahn_alt_def)
  assume "<P',\<sigma>'>\<rightarrow>\<^isub>d\<sigma>''"
  then have b: "\<forall>t' \<in> ic_traces_ending_in_state(\<sigma>'). \<exists>t'' \<in> ic_traces_ending_in_state(\<sigma>''). <P',{t'}>\<rightarrow>\<^isub>a{t''}" by (metis d_kahn_alt_def)
  from a b Kconc1 have "\<forall>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t'' \<in> ic_traces_ending_in_state(\<sigma>''). <P\<parallel>P',{t}>\<rightarrow>\<^isub>a{t''}" by metis
  then show "<P\<parallel>P',\<sigma>>\<rightarrow>\<^isub>d\<sigma>''" by (metis d_kahn_alt_def)
qed
theorem KDconc2: "<P',\<sigma>>\<rightarrow>\<^isub>d\<sigma>' \<Longrightarrow> <P,\<sigma>'>\<rightarrow>\<^isub>d\<sigma>'' \<Longrightarrow> <P\<parallel>P',\<sigma>>\<rightarrow>\<^isub>d\<sigma>''"
  by (metis KDconc1 lang_interleave_commutative)


locale relationships_between_operational_calculi_for_descriptions = small_step_calculi_for_descriptions
begin
lemma plotkin_milner_kahn_descr_rel: "<P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'> \<longleftrightarrow> (\<exists>Q. P -Q\<rightarrow>\<^isub>a P' \<and> <Q,\<sigma>>\<rightarrow>\<^isub>d\<sigma>')"
  by (unfold d_plotkin_def a.a_plotkin_def a.a_milner_def d_kahn_def a_kahn_def, auto)
definition d_state_transformation_function :: "'a description \<Rightarrow> 'a \<Rightarrow> 'a set" ("\<lbrakk>_\<rbrakk>\<^isub>d _" [54] 53) where
  "\<lbrakk>P\<rbrakk>\<^isub>d(\<sigma>) \<equiv> {\<sigma>'. <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'}"
lemma mem_d_state_transformation_function_iff: "\<sigma>' \<in> \<lbrakk>P\<rbrakk>\<^isub>d(\<sigma>) \<longleftrightarrow> <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
  by (auto simp add: d_state_transformation_function_def)
lemma plotkin_milner_state_transformation_descr_rel: "<P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'> \<longleftrightarrow> (\<exists>Q. P -Q\<rightarrow>\<^isub>a P' \<and> \<sigma>' \<in> \<lbrakk>Q\<rbrakk>\<^isub>d(\<sigma>))"
  by (simp add: plotkin_milner_kahn_descr_rel mem_d_state_transformation_function_iff)
lemma d_state_transformation_skip: "\<lbrakk>skip\<rbrakk>\<^isub>d(\<sigma>) = {\<sigma>}"
  by (unfold d_state_transformation_function_def d_kahn_def a_kahn_def, simp add: lang_concat_unit2, auto simp add: trace_can_end_only_in_one_state, metis trivial_ic_trace_ending_in_state)
lemma d_state_transformation_atoms: "\<lbrakk>Rep_atom a\<rbrakk>\<^isub>d(\<sigma>) = a of_state \<sigma>"
proof (unfold d_state_transformation_function_def d_kahn_def a_kahn_def, simp add: atom_applied_to_state_def)
  have "\<forall>\<sigma>'. (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. t' \<in> {t} ; Rep_atom(a)) \<longleftrightarrow> [(\<sigma>, \<sigma>')] \<in> Rep_atom(a)"
  proof (rule allI)
    fix \<sigma>'
    have "(\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. t' \<in> {t} ; Rep_atom(a)) \<longleftrightarrow> (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. \<exists>ta \<in> Rep_atom(a). t' = t @ ta)" by (auto simp add: elem_of_lang_concat)
    also have "\<dots> \<longleftrightarrow> (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. \<exists>\<sigma>1 \<sigma>2. [(\<sigma>1,\<sigma>2)] \<in> Rep_atom(a) \<and> t' = t @ [(\<sigma>1,\<sigma>2)])" by (metis trace_of_atom_characterization)
    also have "\<dots> \<longleftrightarrow> (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. \<exists>\<sigma>1 \<sigma>2. [(\<sigma>1,\<sigma>2)] \<in> Rep_atom(a) \<and> t' = t @ [(\<sigma>1,\<sigma>2)] \<and> snd(last t') = \<sigma>')" by (auto simp add: last_state_of_trace_ending_in_state)
    also have "\<dots> \<longleftrightarrow> (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. \<exists>\<sigma>1. [(\<sigma>1,\<sigma>')] \<in> Rep_atom(a) \<and> t' = t @ [(\<sigma>1,\<sigma>')])" by (metis snd_conv snoc_eq_iff_butlast)
    also have "\<dots> \<longleftrightarrow> (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. \<exists>\<sigma>1. [(\<sigma>1,\<sigma>')] \<in> Rep_atom(a) \<and> t' = t @ [(\<sigma>1,\<sigma>')] \<and> t \<noteq> [] \<and> [(\<sigma>1,\<sigma>')] \<noteq> [] \<and> ic(t @ [(\<sigma>1,\<sigma>')]))" by (auto simp add: ic_traces_ending_in_state_def)
    also have "\<dots> \<longleftrightarrow> (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. \<exists>\<sigma>1. [(\<sigma>1,\<sigma>')] \<in> Rep_atom(a) \<and> t' = t @ [(\<sigma>1,\<sigma>')] \<and> t \<noteq> [] \<and> [(\<sigma>1,\<sigma>')] \<noteq> [] \<and> ic(t @ [(\<sigma>1,\<sigma>')]) \<and> snd (last t) = fst(hd [(\<sigma>1,\<sigma>')]))" by (metis consistent_append_nonempties)
    also have "\<dots> \<longleftrightarrow> (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. \<exists>\<sigma>1. [(\<sigma>1,\<sigma>')] \<in> Rep_atom(a) \<and> t' = t @ [(\<sigma>1,\<sigma>')] \<and> t \<noteq> [] \<and> [(\<sigma>1,\<sigma>')] \<noteq> [] \<and> ic(t @ [(\<sigma>1,\<sigma>')]) \<and> \<sigma> = \<sigma>1)" by (auto simp add: last_state_of_trace_ending_in_state)
    also have "\<dots> \<longleftrightarrow> (\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. [(\<sigma>,\<sigma>')] \<in> Rep_atom(a) \<and> t' = t @ [(\<sigma>,\<sigma>')])" by (auto simp add: ic_traces_ending_in_state_def)
    also have "\<dots> \<longleftrightarrow> [(\<sigma>, \<sigma>')] \<in> Rep_atom(a)" by (metis ic_traces_ending_in_state_snoc)
    finally show "(\<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. t' \<in> {t} ; Rep_atom(a)) \<longleftrightarrow> [(\<sigma>, \<sigma>')] \<in> Rep_atom(a)" by simp
  qed
  then show "{\<sigma>'. \<exists>t\<in>ic_traces_ending_in_state \<sigma>. \<exists>t'\<in>ic_traces_ending_in_state \<sigma>'. t' \<in> {t} ; Rep_atom(a)} = {\<sigma>'. [(\<sigma>, \<sigma>')] \<in> Rep_atom(a)}" by auto
qed
lemma plotkin_kahn_descr_rel_aux: "<P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'> \<Longrightarrow> <P',\<sigma>'>\<rightarrow>\<^isub>d\<sigma>'' \<Longrightarrow> <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>''"
proof -
  assume a: "<P,\<sigma>>\<rightarrow>\<^isub>d<P',\<sigma>'>"
  hence "\<exists>t \<in> ic_traces_ending_in_state(\<sigma>). \<exists>t' \<in> ic_traces_ending_in_state(\<sigma>'). \<exists>Q. Q \<in> Actions \<and> P \<supseteq> Q;P' \<and> {t};Q \<supseteq> {t'}" by (simp add: d_plotkin_def a.a_plotkin_def)
  then obtain t t' Q where a1: "t \<in> ic_traces_ending_in_state(\<sigma>)" and a2: "t' \<in> ic_traces_ending_in_state(\<sigma>')" and a3: "P \<supseteq> Q;P'" and "{t};Q \<supseteq> {t'}" by blast
  hence "\<exists>tq \<in> Q. t@tq = t'" by (metis elem_of_lang_concat insert_subset singletonE)
  then obtain tq where a4: "tq \<in> Q" and a5: "t@tq = t'" by blast
  assume b: "<P',\<sigma>'>\<rightarrow>\<^isub>d\<sigma>''"
  hence "\<forall>t' \<in> ic_traces_ending_in_state(\<sigma>'). \<exists>t'' \<in> ic_traces_ending_in_state(\<sigma>''). {t''} \<subseteq> {t'};P'" by (simp add: d_kahn_alt_def a_kahn_def)
  with a2 have "\<exists>t'' \<in> ic_traces_ending_in_state(\<sigma>''). {t''} \<subseteq> {t'};P'" by metis
  then obtain t'' where b1: "t'' \<in> ic_traces_ending_in_state(\<sigma>'')" and "{t''} \<subseteq> {t'};P'" by blast
  hence "\<exists>tp' \<in> P'. t'' = t'@tp'" by (metis elem_of_lang_concat insert_subset singletonE)
  then obtain tp' where b2: "tp' \<in> P'" and b3: "t'' = t'@tp'" by blast
  show "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>''"
  proof -
    have "\<exists>t\<in>ic_traces_ending_in_state(\<sigma>). \<exists>t''\<in>ic_traces_ending_in_state(\<sigma>''). <P,{t}>\<rightarrow>\<^isub>a{t''}"
    proof -
      from a3 a4 b2 have "tq@tp' \<in> P" by (metis elem_of_lang_concat set_rev_mp)
      hence "{t};P \<supseteq> {t@tq@tp'}" by (metis bot_least elem_of_lang_concat insertI1 insert_subset)
      with a5 b3 have "{t};P \<supseteq> {t''}" by (metis append_assoc)
      with a1 b1 a_kahn_def a b d_plotkin_def show "\<exists>t\<in>ic_traces_ending_in_state(\<sigma>). \<exists>t''\<in>ic_traces_ending_in_state(\<sigma>''). <P,{t}>\<rightarrow>\<^isub>a{t''}" by metis
    qed
    with a b show "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>''" by (metis d_kahn_def)
  qed
qed
lemma plotkin_kahn_descr_rel: "<P,\<sigma>>\<rightarrow>\<^isub>d\<^sup>\<star><skip,\<sigma>'> \<Longrightarrow> <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
proof (unfold d_plotkin_refl_trans_def)
  assume "\<exists>n. <P,\<sigma>>\<rightarrow>\<^isub>d[n]<skip,\<sigma>'>"
  then obtain n where a: " <P,\<sigma>>\<rightarrow>\<^isub>d[n]<skip,\<sigma>'>" by blast
  have "\<forall>P \<sigma> \<sigma>'. <P,\<sigma>>\<rightarrow>\<^isub>d[n]<skip,\<sigma>'> \<longrightarrow> <P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by (induct n, simp, metis KDskip, auto simp add: plotkin_kahn_descr_rel_aux)
  with a show "<P,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by metis
qed
end

locale small_step_calculi_for_commands = small_step_calculi_for_descriptions
begin
definition c_plotkin :: "'a command \<Rightarrow> 'a \<Rightarrow> 'a command \<Rightarrow> 'a \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>c<_,_>" [54,54,54,54] 53) where
  "<C,\<sigma>>\<rightarrow>\<^isub>c<C',\<sigma>'> \<equiv> <trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d<trace_set_of_command(C'),\<sigma>'>"
lemma PCatom: "a \<in> AtomicOperations \<Longrightarrow> \<sigma>' \<in> a of_state \<sigma> \<Longrightarrow> <{[a]},\<sigma>>\<rightarrow>\<^isub>c<skip,\<sigma>'>"
  by (metis PDatom atom_homomorphism c_plotkin_def skip_homomorphism)
lemma PCseq1: "<skip;C,\<sigma>>\<rightarrow>\<^isub>c<C,\<sigma>>"
  by (auto simp add: PDseq1 c_plotkin_def skip_homomorphism seq_homomorphism)
lemma PCseq2: "<C,\<sigma>>\<rightarrow>\<^isub>c<C',\<sigma>'> \<Longrightarrow> <C;C'',\<sigma>>\<rightarrow>\<^isub>c<C';C'',\<sigma>'>"
  by (unfold c_plotkin_def, metis (no_types) PDseq2 seq_homomorphism)
lemma PCchoice: "C \<in> Y \<Longrightarrow> <\<Union>Y,\<sigma>>\<rightarrow>\<^isub>c<C,\<sigma>>"
proof -
  assume a3: "C \<in> Y"
  with a3 have "trace_set_of_command(C) \<in> {P. \<exists>C \<in> Y. P = trace_set_of_command(C)}" by auto
  hence "<\<Union>{P. \<exists>C \<in> Y. P = trace_set_of_command(C)},\<sigma>>\<rightarrow>\<^isub>d<trace_set_of_command(C),\<sigma>>" by (rule PDchoice)
  hence b3: "<trace_set_of_command(\<Union>Y),\<sigma>>\<rightarrow>\<^isub>d<trace_set_of_command(C),\<sigma>>" by (metis choice_homomorphism)
  show "<\<Union>Y,\<sigma>>\<rightarrow>\<^isub>c<C,\<sigma>>" by (metis b3 c_plotkin_def)
qed
lemma PCiter1: "<C\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>c<skip,\<sigma>>"
  by (unfold c_plotkin_def, metis PDiter1 kleene_homomorphism skip_homomorphism)
lemma PCiter2: "<C\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>c<C;C\<^sup>\<star>,\<sigma>>"
  by (auto simp add: PDiter2 c_plotkin_def kleene_homomorphism skip_homomorphism seq_homomorphism)
lemma PCconc1: "<skip\<parallel>C,\<sigma>>\<rightarrow>\<^isub>c<C,\<sigma>>" 
  by (auto simp add: PDconc1 c_plotkin_def skip_homomorphism conc_homomorphism)
lemma PCconc2: "<C\<parallel>skip,\<sigma>>\<rightarrow>\<^isub>c<C,\<sigma>>"
  by (auto simp add: PDconc2 c_plotkin_def skip_homomorphism conc_homomorphism)
lemma PCconc3: "<C,\<sigma>>\<rightarrow>\<^isub>c<C',\<sigma>'> \<Longrightarrow> <C\<parallel>C'',\<sigma>>\<rightarrow>\<^isub>c<C'\<parallel>C'',\<sigma>'>"
  by (unfold c_plotkin_def, metis (no_types) PDconc3 conc_homomorphism)
lemma PCconc4: "<C,\<sigma>>\<rightarrow>\<^isub>c<C',\<sigma>'> \<Longrightarrow> <C''\<parallel>C,\<sigma>>\<rightarrow>\<^isub>c<C''\<parallel>C',\<sigma>'>"
  by (unfold c_plotkin_def, metis (no_types) PDconc4 conc_homomorphism)
fun c_plotkin_iter :: "'a command \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a command \<Rightarrow> 'a \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>c[_]<_,_>" [54,54,54,54,54] 53) where
   "<C,\<sigma>>\<rightarrow>\<^isub>c[0]<C',\<sigma>'> = (C = C' \<and> \<sigma> = \<sigma>')"
 | "<C,\<sigma>>\<rightarrow>\<^isub>c[Suc n]<C',\<sigma>'> = (\<exists>C'' \<sigma>''. <C,\<sigma>>\<rightarrow>\<^isub>c<C'',\<sigma>''> \<and> <C'',\<sigma>''>\<rightarrow>\<^isub>c[n]<C',\<sigma>'>)"
definition c_plotkin_refl_trans :: "'a command \<Rightarrow> 'a \<Rightarrow> 'a command \<Rightarrow> 'a \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>c\<^sup>\<star><_,_>" [54,54,54,54] 53) where
   "<C,\<sigma>>\<rightarrow>\<^isub>c\<^sup>\<star><C',\<sigma>'> \<equiv> (\<exists>n. <C,\<sigma>>\<rightarrow>\<^isub>c[n]<C',\<sigma>'>)"
lemma c_plotkin_iter_d_plotkin_iter_relationship: "\<And>C \<sigma> C' \<sigma>'. <C,\<sigma>>\<rightarrow>\<^isub>c[n]<C',\<sigma>'> \<Longrightarrow> <trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d[n]<trace_set_of_command(C'),\<sigma>'>"
proof (induct n, simp, simp)
  fix n C \<sigma> C' \<sigma>'
  assume IH: "\<And>C \<sigma> C' \<sigma>'. <C,\<sigma>>\<rightarrow>\<^isub>c[n]<C',\<sigma>'> \<Longrightarrow> <trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d[n]<trace_set_of_command(C'),\<sigma>'>"
  assume "\<exists>C'' \<sigma>''. <C,\<sigma>>\<rightarrow>\<^isub>c<C'',\<sigma>''> \<and> <C'',\<sigma>''>\<rightarrow>\<^isub>c[n]<C',\<sigma>'>"
  then obtain C'' \<sigma>'' where a1: "<C,\<sigma>>\<rightarrow>\<^isub>c<C'',\<sigma>''>" and a2: "<C'',\<sigma>''>\<rightarrow>\<^isub>c[n]<C',\<sigma>'>" by blast
  from a2 have a3: "<trace_set_of_command(C''),\<sigma>''>\<rightarrow>\<^isub>d[n]<trace_set_of_command(C'),\<sigma>'>" by (rule IH)
  show "\<exists>P'' \<sigma>''. <trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d<P'',\<sigma>''> \<and> <P'',\<sigma>''>\<rightarrow>\<^isub>d[n]<trace_set_of_command(C'),\<sigma>'>" by (metis a1 a3 c_plotkin_def)
qed
lemma c_plotkin_refl_trans_d_plotkin_refl_trans_relationship: "<C,\<sigma>>\<rightarrow>\<^isub>c\<^sup>\<star><C',\<sigma>'> \<Longrightarrow> <trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d\<^sup>\<star><trace_set_of_command(C'),\<sigma>'>"
  by (unfold c_plotkin_refl_trans_def d_plotkin_refl_trans_def, metis c_plotkin_iter_d_plotkin_iter_relationship)

definition  c_milner :: "'a command \<Rightarrow> 'a command \<Rightarrow> 'a command \<Rightarrow> bool" ("_ -_\<rightarrow>\<^isub>c _" [54,54,54] 53) where
  "C -C'\<rightarrow>\<^isub>c C'' \<equiv> trace_set_of_command(C) -trace_set_of_command(C')\<rightarrow>\<^isub>a trace_set_of_command(C'')"
lemma MCatom: "a \<in> AtomicOperations \<Longrightarrow> {[a]} -{[a]}\<rightarrow>\<^isub>c skip"
proof (unfold c_milner_def)
  assume "a \<in> AtomicOperations"
  hence "Rep_atom(a) \<in> Actions" by (unfold Actions_def, auto)
  thus "abstract_small_step_calculi.a_milner Actions (trace_set_of_command {[a]}) (trace_set_of_command {[a]}) (trace_set_of_command skip)" by (metis a.Maction atom_homomorphism skip_homomorphism)
qed
lemma MCseq1: "skip;C -skip\<rightarrow>\<^isub>c C"
  by (auto simp add: a.Mseq1 c_milner_def skip_homomorphism seq_homomorphism)
lemma MCseq2: "C -C'\<rightarrow>\<^isub>c C'' \<Longrightarrow> C;C1 -C'\<rightarrow>\<^isub>c C'';C1"
  by (unfold c_milner_def, metis (no_types) a.Mseq2 seq_homomorphism)
lemma MCchoice: "C \<in> Y \<Longrightarrow> \<Union>Y -skip\<rightarrow>\<^isub>c C"
proof -
  assume "C \<in> Y"
  hence "trace_set_of_command(C) \<in> {P. \<exists>C \<in> Y. P = trace_set_of_command(C)}" by auto
  hence "\<Union>{P. \<exists>C \<in> Y. P = trace_set_of_command(C)} -skip\<rightarrow>\<^isub>a trace_set_of_command(C)" by (rule a.Mchoice)
  hence "trace_set_of_command(\<Union>Y) -trace_set_of_command(skip)\<rightarrow>\<^isub>a trace_set_of_command(C)" by (metis choice_homomorphism skip_homomorphism)
  thus "\<Union>Y -skip\<rightarrow>\<^isub>c C" by (metis c_milner_def)
qed
lemma MCiter1: "C\<^sup>\<star> -skip\<rightarrow>\<^isub>c skip"
  by (unfold c_milner_def, metis a.Miter1 kleene_homomorphism skip_homomorphism)
lemma MCiter2: "C\<^sup>\<star> -skip\<rightarrow>\<^isub>c C;C\<^sup>\<star>"
  by (auto simp add: a.Miter2 c_milner_def kleene_homomorphism skip_homomorphism seq_homomorphism)
lemma MCconc1: "skip\<parallel>C -skip\<rightarrow>\<^isub>c C"
  by (auto simp add: a.Mconc1 c_milner_def skip_homomorphism conc_homomorphism)
lemma MCconc2: "C\<parallel>skip -skip\<rightarrow>\<^isub>c C"
  by (auto simp add: a.Mconc2 c_milner_def skip_homomorphism conc_homomorphism)
lemma MCconc3: "C -C'\<rightarrow>\<^isub>c C'' \<Longrightarrow> C\<parallel>C1 -C'\<rightarrow>\<^isub>c C''\<parallel>C1"
  by (unfold c_milner_def, metis (no_types) a.Mconc3 conc_homomorphism)
lemma MCconc4: "C -C'\<rightarrow>\<^isub>c C'' \<Longrightarrow> C1\<parallel>C -C'\<rightarrow>\<^isub>c C1\<parallel>C''"
  by (unfold c_milner_def, metis (no_types) a.Mconc4 conc_homomorphism)
end

definition c_kahn :: "'a command \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> bool" ("<_,_>\<rightarrow>\<^isub>c_" [54,54,54] 53) where
  "<C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>' \<equiv> <trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
theorem KCatom: "\<sigma>' \<in> a of_state \<sigma> \<Longrightarrow> <{[a]},\<sigma>>\<rightarrow>\<^isub>c\<sigma>'"
  by (metis (full_types) KDatom atom_homomorphism c_kahn_def)
theorem KCskip: "<skip,\<sigma>>\<rightarrow>\<^isub>c\<sigma>"
  by (metis KDskip c_kahn_def skip_homomorphism)
theorem KCseq: "<C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>' \<Longrightarrow> <C',\<sigma>'>\<rightarrow>\<^isub>c\<sigma>'' \<Longrightarrow> <C;C',\<sigma>>\<rightarrow>\<^isub>c\<sigma>''"
  by (auto simp add: KDseq c_kahn_def seq_homomorphism)
theorem KCchoice: "C \<in> Y \<Longrightarrow> <C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>' \<Longrightarrow> <\<Union>Y,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'"
proof -
  assume a1: "C \<in> Y" and a2: "<C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'"
  with a1 have b1: "trace_set_of_command(C) \<in> {P. \<exists>C \<in> Y. P = trace_set_of_command(C)}" by auto
  from a2 have b2: "<trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by (simp add: c_kahn_def)
  from b1 b2 have "<\<Union>{P. \<exists>C \<in> Y. P = trace_set_of_command(C)},\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by (rule KDchoice)
  hence "<trace_set_of_command(\<Union>Y),\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by (metis choice_homomorphism)
  thus "<\<Union>Y,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'" by (metis c_kahn_def)
qed
theorem KCiter1: "<C\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>c\<sigma>"
  by (metis KDiter1 c_kahn_def kleene_homomorphism)
theorem KCiter2: "<C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>' \<Longrightarrow> <C\<^sup>\<star>,\<sigma>'>\<rightarrow>\<^isub>c\<sigma>'' \<Longrightarrow> <C\<^sup>\<star>,\<sigma>>\<rightarrow>\<^isub>c\<sigma>''"
  by (metis KDiter2 c_kahn_def kleene_homomorphism)
theorem KCconc1: "<C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>' \<Longrightarrow> <C',\<sigma>'>\<rightarrow>\<^isub>c\<sigma>'' \<Longrightarrow> <C\<parallel>C',\<sigma>>\<rightarrow>\<^isub>c\<sigma>''"
  by (auto simp add: KDconc1 c_kahn_def conc_homomorphism)
theorem KCconc2: "<C',\<sigma>>\<rightarrow>\<^isub>c\<sigma>' \<Longrightarrow> <C,\<sigma>'>\<rightarrow>\<^isub>c\<sigma>'' \<Longrightarrow> <C\<parallel>C',\<sigma>>\<rightarrow>\<^isub>c\<sigma>''"
  by (auto simp add: KDconc2 c_kahn_def conc_homomorphism)

locale relationships_between_operational_calculi_for_commands = relationships_between_operational_calculi_for_descriptions + small_step_calculi_for_commands
begin
lemma plotkin_milner_kahn_comm_rel: "<C,\<sigma>>\<rightarrow>\<^isub>c<C',\<sigma>'> \<longleftrightarrow> (\<exists>C''. C -C''\<rightarrow>\<^isub>c C' \<and> <C'',\<sigma>>\<rightarrow>\<^isub>c\<sigma>')"
proof -
  have "<C,\<sigma>>\<rightarrow>\<^isub>c<C',\<sigma>'> \<longleftrightarrow> <trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d<trace_set_of_command(C'),\<sigma>'>" by (simp add: c_plotkin_def)
  also have "\<dots> \<longleftrightarrow> (\<exists>Q. trace_set_of_command(C) -Q\<rightarrow>\<^isub>a trace_set_of_command(C') \<and> <Q,\<sigma>>\<rightarrow>\<^isub>d\<sigma>')" by (simp add: plotkin_milner_kahn_descr_rel)
  also have "\<dots> \<longleftrightarrow> (\<exists>C''. C -C''\<rightarrow>\<^isub>c C' \<and> <C'',\<sigma>>\<rightarrow>\<^isub>c\<sigma>')"
  proof (rule iffI)
    assume "\<exists>Q. trace_set_of_command(C) -Q\<rightarrow>\<^isub>a trace_set_of_command(C') \<and> <Q,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'"
    then obtain Q where a1: "trace_set_of_command(C) -Q\<rightarrow>\<^isub>a trace_set_of_command(C')" and a2: "<Q,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by blast
    from a1 have "Q \<in> {skip} \<union> Rep_atom ` AtomicOperations" by (metis a.a_milner_def Actions_def)
    hence "Q = skip \<or> Q \<in> Rep_atom ` AtomicOperations" by auto
    moreover {
      assume b: "Q = skip"
      from a1 b have b1: "C -skip\<rightarrow>\<^isub>c C'" by (auto simp add: skip_homomorphism c_milner_def)
      from a2 b have b2: "<skip,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'" by (metis skip_homomorphism c_kahn_def)
      from b1 b2 have "\<exists>C''. C -C''\<rightarrow>\<^isub>c C' \<and> <C'',\<sigma>>\<rightarrow>\<^isub>c\<sigma>'" by auto
    } moreover {
      assume b: "Q \<in> Rep_atom ` AtomicOperations"
      with a1 have b1: "C -{[Abs_atom Q]}\<rightarrow>\<^isub>c C'" by (metis Rep_atom_inverse atom_homomorphism c_milner_def image_iff)
      from a2 b have b2: "<{[Abs_atom Q]},\<sigma>>\<rightarrow>\<^isub>c\<sigma>'" by (metis atom_homomorphism c_kahn_def Rep_atom_inverse image_iff)
      from b1 b2 have "\<exists>C''. C -C''\<rightarrow>\<^isub>c C' \<and> <C'',\<sigma>>\<rightarrow>\<^isub>c\<sigma>'" by auto
    }
    ultimately show "\<exists>C''. C -C''\<rightarrow>\<^isub>c C' \<and> <C'',\<sigma>>\<rightarrow>\<^isub>c\<sigma>'" by (rule disjE) 
  next
    assume "\<exists>C''. C -C''\<rightarrow>\<^isub>c C' \<and> <C'',\<sigma>>\<rightarrow>\<^isub>c\<sigma>'"
    thus "\<exists>Q. trace_set_of_command(C) -Q\<rightarrow>\<^isub>a trace_set_of_command(C') \<and> <Q,\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by (metis c_kahn_def c_milner_def)
  qed
  finally show ?thesis by simp
qed
definition c_state_transformation_function :: "'a command \<Rightarrow> 'a \<Rightarrow> 'a set" ("\<lbrakk>_\<rbrakk>\<^isub>c _" [54] 53) where
  "\<lbrakk>C\<rbrakk>\<^isub>c(\<sigma>) \<equiv> {\<sigma>'. <C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'}"
lemma mem_d_state_transformation_function_iff: "\<sigma>' \<in> \<lbrakk>C\<rbrakk>\<^isub>c(\<sigma>) \<longleftrightarrow> <C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'"
  by (auto simp add: c_state_transformation_function_def)
lemma plotkin_milner_state_transformation_comm_rel: "<C,\<sigma>>\<rightarrow>\<^isub>c<C',\<sigma>'> \<longleftrightarrow> (\<exists>C''. C -C''\<rightarrow>\<^isub>c C' \<and> \<sigma>' \<in> \<lbrakk>C''\<rbrakk>\<^isub>c(\<sigma>))"
  by (simp add: plotkin_milner_kahn_comm_rel mem_d_state_transformation_function_iff)
lemma c_state_transformation_function_d_state_transformation_function_rel: "\<lbrakk>C\<rbrakk>\<^isub>c(\<sigma>) = \<lbrakk>trace_set_of_command(C)\<rbrakk>\<^isub>d(\<sigma>)"
  by (unfold c_state_transformation_function_def d_state_transformation_function_def c_kahn_def, auto)
lemma plotkin_kahn_comm_rel: "<C,\<sigma>>\<rightarrow>\<^isub>c\<^sup>\<star><skip,\<sigma>'> \<Longrightarrow> <C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'"
proof -
  assume "<C,\<sigma>>\<rightarrow>\<^isub>c\<^sup>\<star><skip,\<sigma>'>"
  hence "<trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d\<^sup>\<star><trace_set_of_command(skip),\<sigma>'>" by (metis c_plotkin_refl_trans_d_plotkin_refl_trans_relationship)
  hence "<trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d\<^sup>\<star><skip,\<sigma>'>" by (metis skip_homomorphism)
  hence "<trace_set_of_command(C),\<sigma>>\<rightarrow>\<^isub>d\<sigma>'" by (metis plotkin_kahn_descr_rel)
  thus "<C,\<sigma>>\<rightarrow>\<^isub>c\<sigma>'" by (metis c_kahn_def)
qed
end

end
